@@ -1,6 +1,6 @@
 import numpy as np
 from xrmreader.reader import read_txrm_iterable, read_metadata
-from scipy.ndimage import shift
+from scipy.ndimage import shift, convolve
 from tqdm import tqdm
 from skimage.measure import block_reduce
 from skimage.transform.pyramids import pyramid_reduce
@@ -76,7 +76,7 @@ def downsample(projections, spatial_factor: int, angular_factor: int):
     # rather than leaving out every n-th value, compute the mean value over a fixed sized window for reduction
     # projections = block_reduce(projections, block_size=(1, spatial_factor, spatial_factor), func=np.mean)
     if spatial_factor > 1:
-        projections = pyramid_reduce(projections, downscale=spatial_factor)
+        projections = pyramid_reduce(projections, downscale=spatial_factor, preserve_range=True)
 
     return projections
 
@@ -88,8 +88,7 @@ def truncation_correction(projections, extension_fraction=0.1):
     :param extension_fraction: faction of image width to be added to both sides
     :return: extended projections
     '''
-    # # todo: use simple numpy solution or implement https://pubmed.ncbi.nlm.nih.gov/10659736/ ??
-    # # this is not noise-maintaining because no copy image content
+    # todo: use simple numpy solution or implement https://pubmed.ncbi.nlm.nih.gov/10659736/ ??
     num_images, img_height, img_width = projections.shape
     n_ext = int(img_width * extension_fraction)
     projection = np.pad(projections, pad_width=((0, 0), (0, 0), (n_ext, n_ext)), mode='symmetric')
@@ -99,11 +98,53 @@ def truncation_correction(projections, extension_fraction=0.1):
     return projection
 
 
+def insert_quantum_noise(p_A, p_air, p_air_smooth, a=1.):
+    '''Simulates quantum noise following eq. (6) and (8) from "Lifeng Yu - Development and Validation of a Practical
+    Lower-Dose-Simulation Tool for Optimizing Computed Tomography Scan Protocols".
+
+    :param p_A: projection image with object
+    :param p_air: reference image without object to calculate N_0A
+    :param p_air_smooth: smoothed reference projection to properly calculate the quantum noise
+    :param a: N_0B = a * N_0A, i.e. "dose(P_B)/dose(P_A)" 0 < a <= 1
+    :return: Image with inserted quantum noise P_B
+    '''
+    if a == 1.:
+        return p_A
+    else:
+        p_air = np.array(p_air)
+        p_air_norm = p_air / p_air_smooth
+        p_air_log = -1 * np.log(p_air_norm)
+
+        mu = 0
+        sigma = 1
+        x = np.random.normal(mu, sigma, np.shape(p_A))
+        n_0A = 1 / np.var(np.exp(-p_air_log))
+        p_B = p_A + np.sqrt(((1-a) / a) * (np.exp(p_A) / n_0A)) * x
+        return p_B
+
+
+def smooth_reference(reference, kernel_size=15):
+    '''Smooth the reference projection with a large kernel to receive the global intensity distribution of the incident
+    beam. Pretty long runtime, so just do this once.
+
+    :param reference: reference projection
+    :param kernel_size: kernel size
+    :return: Smoothed reference image
+    '''
+    reference = np.array(reference)
+    kernel = np.ones((kernel_size, kernel_size)) / np.sum(np.ones((kernel_size, kernel_size)))
+    reference_smooth = convolve(reference, kernel, mode='reflect')
+    # reference_smooth = convolve2d(reference, kernel, mode='same', boundary='symm')
+    return reference_smooth
+
+
 def read_and_preprocess_txrm(projections_file, downsample_factor=1, angular_factor=1, do_truncation_correction=True,
-                             extension_fraction=0.1, datatype=np.float32, keep_rows=None):
+                             extension_fraction=0.1, dose_reduction=1., datatype=np.float32, keep_rows=None):
     '''Preprocess the raw images, i.e. remove shifts, transform into line integral domain.
     Projections are extended for truncation correction if needed.
 
+    :param keep_rows: bool type array of same length as image height indicating which detector rows to keep
+    :param dose_reduction: determines the dose relative to the high dose acquisition -> Poisson noise is inserted
     :param projections_file: Path to txrm projection file, images are loaded on-the-fly to avoid high memory consumption
     :param downsample_factor: factor by which the projections are downsampled spatially; 1 means no downsampling
     :param angular_factor: factor by which the number of projections is downsampled; 1 means no downsampling
@@ -118,9 +159,11 @@ def read_and_preprocess_txrm(projections_file, downsample_factor=1, angular_fact
     x_shifts = metadata['x-shifts'][::angular_factor]
     y_shifts = metadata['y-shifts'][::angular_factor]
     reference = metadata['reference']
-    keep_rows = keep_rows.astype(np.bool)
+    if not dose_reduction == 1.:
+        reference_smooth = smooth_reference(reference)
     if keep_rows is not None:
-        projection_height = int(np.sum(keep_rows) / downsample_factor)
+        keep_rows = keep_rows.astype(np.bool)
+        projection_height = int(np.ceil(np.sum(keep_rows) / downsample_factor))
     else:
         projection_height = int(np.ceil(float(metadata['image_height']) / downsample_factor))
     print('Preprocessing projections')
@@ -143,16 +186,16 @@ def read_and_preprocess_txrm(projections_file, downsample_factor=1, angular_fact
             zip(read_txrm_iterable(str(projections_file), slice_range=(
             (0, num_images, angular_factor), (0, metadata['image_height'], 1), (0, metadata['image_width'], 1))),
                 x_shifts, y_shifts)), total=np.ceil(num_images / angular_factor)):
-        # if i % 10 == 0:
-        #     print(f'{i + 1} / {int(np.ceil(num_images / angular_factor))}')
         projection = np.expand_dims(projection, 0)
         projection = divide_by_reference(projection, reference)
         projection = revert_shifts(projection, x_shift, y_shift)
         if keep_rows is not None:
             projection = projection[:, keep_rows, :]
+        projection = negative_logarithm(projection)
+        if not dose_reduction == 1.:
+            projection = insert_quantum_noise(projection, reference, reference_smooth, dose_reduction)
         if downsample_factor != 1 or angular_factor != 1:
             projection = downsample(projection, downsample_factor, angular_factor)
-        projection = negative_logarithm(projection)
         if do_truncation_correction:
             projection = truncation_correction(projection, extension_fraction=extension_fraction)
         projections[i, :, :] = np.squeeze(projection)
@@ -161,18 +204,23 @@ def read_and_preprocess_txrm(projections_file, downsample_factor=1, angular_fact
 
 
 def read_single_projection(projections_file, downsample_factor=1, angular_factor=1, do_truncation_correction=True,
-                             extension_fraction=0.1, datatype=np.float32, keep_rows=None):
+                             extension_fraction=0.1, dose_reduction=1., datatype=np.float32, keep_rows=None):
     metadata = read_metadata(str(projections_file))
     reference = metadata['reference']
-    keep_rows = keep_rows.astype(np.bool)
+    if not dose_reduction == 1.:
+        reference_smooth = smooth_reference(reference)
+    if keep_rows is not None:
+        keep_rows = keep_rows.astype(np.bool)
     for projection in read_txrm_iterable(str(projections_file)):
         projection = np.expand_dims(projection, 0)
         projection = divide_by_reference(projection, reference)
         if keep_rows is not None:
             projection = projection[:, keep_rows, :]
+        projection = negative_logarithm(projection)
+        if not dose_reduction == 1.:
+            projection = insert_quantum_noise(projection, reference, reference_smooth, dose_reduction)
         if downsample_factor != 1 or angular_factor != 1:
             projection = downsample(projection, downsample_factor, angular_factor)
-        projection = negative_logarithm(projection)
         if do_truncation_correction:
             projection = truncation_correction(projection, extension_fraction=extension_fraction)
         break
